# -*- coding:utf-8 -*-

import operator

import numpy as np

__copyright__ = 'Copyright 2016'
__author__ = u'Bsc. Manuel Aguado Martínez'


def calculate_eer_step_by_step(genuine_match_scores, imp_match_scores,
                               match_thr_step, hist=False):
    """Calculates the Equal Error Rate measure increasing in each step
    the matching threshold in a given value.

    Keyword Arguments:
    genuine_match_scores -- The score of each genuine match
    imp_match_scores -- The score of each impostor match
    match_thr_step -- The value in which increase the threshold at each step
    hist -- Indicates whether the impostor_match_scores came in histogram
            format. Avalaible only for integer matching thresholds step.

    :returns a tuple with the folling format: (thresholds, false match rates,
        false non match rates, EER value)
    """
    # TODO Improve the efficiency of this algorithm using numpy array from
    # the beginning for error lists

    match_thr = 0
    if hist:
        maximum_thr = max(max(genuine_match_scores), len(imp_match_scores))
    else:
        maximum_thr = max(genuine_match_scores)

    total_true_match = float(len(genuine_match_scores))
    if hist:
        total_false_match = float(sum(imp_match_scores))
        imp_match_scores = np.cumsum(imp_match_scores)
    else:
        total_false_match = float(len(imp_match_scores))

    thresholds = []
    false_match_rates = []
    false_non_match_rates = []

    while match_thr <= maximum_thr:
        if hist:
            if match_thr < len(imp_match_scores):
                false_match = total_false_match - imp_match_scores[match_thr]
            else:
                false_match = 0
        else:
            # List convertion for Python3 compatibility
            false_match = len(list(filter(lambda s: s > match_thr,
                                     imp_match_scores)))

        false_match_rates.append(false_match / total_false_match)

        # List convertion for Python3 compatibility
        false_non_match = len(list(filter(lambda s: s <= match_thr,
                                     genuine_match_scores)))
        false_non_match_rates.append(false_non_match / total_true_match)

        thresholds.append(match_thr)
        match_thr += match_thr_step

    false_match_rates = np.array(false_match_rates)
    false_non_match_rates = np.array(false_non_match_rates)
    index = np.argmin(abs(false_match_rates - false_non_match_rates))
    eer = abs(false_non_match_rates[index] + false_match_rates[index]) / 2.0

    return thresholds, false_match_rates, false_non_match_rates, eer


def calculate_eer(genuine_match_scores, impostor_match_scores):
    """Calculates the Equal Error Rate measure

    Keyword Arguments:
    genuine_match_scores -- The score of each genuine match
    impostor_match_scores -- The score of each impostor match

    :returns a tuple with the folling format: (thresholds, false match rates,
        false non match rates, EER value)
    """
    total_true_match = len(genuine_match_scores)
    total_false_match = len(impostor_match_scores)
    total = total_true_match + total_false_match

    genuine_match_scores = zip(genuine_match_scores, [1] * total_true_match)    
    genuine_match_scores = list(genuine_match_scores) # Python3 compatibility

    impostor_match_scores = zip(impostor_match_scores, [0] * total_false_match)    
    impostor_match_scores = list(impostor_match_scores) # Python3 compatibility

    scores = np.array(sorted(genuine_match_scores + impostor_match_scores,
                             key=operator.itemgetter(0)))
    sum_true = np.cumsum(scores[:, 1])
    sum_false = total_false_match - (np.arange(0, total) - sum_true)

    false_match_rates = np.ones(total + 1)
    false_non_match_rates = np.zeros(total + 1)

    false_non_match_rates[1:] = sum_true / total_true_match
    false_match_rates[1:] = sum_false / total_false_match

    thresholds = np.zeros((total + 1))
    thresholds[1:] = scores[:, 0]

    index = np.argmin(abs(false_match_rates - false_non_match_rates))
    eer = abs(false_non_match_rates[index] + false_match_rates[index]) / 2.0

    return thresholds, false_match_rates, false_non_match_rates, eer




# -*- coding:utf-8 -*-

import argparse
from os.path import join

import matplotlib.pyplot as plt
import matplotlib.font_manager as font
import numpy as np

from stats import calculate_eer, calculate_eer_step_by_step

__copyright__ = 'Copyright 2016'
__author__ = u'Bsc. Manuel Aguado Martínez'


def __get_score(line):
    """Get the score value from an input score file line

    Keyword Arguments:
    line -- An input score file line
    """
    sline = line.strip().split(' ')
    return float(sline[-1])


def get_eer_info():
    ap = argparse.ArgumentParser()
    ap.add_argument("-p", "--path", required=True, help="path to match files")
    ap.add_argument("-i", "--impostor_match_file_names", required=True,
                    help="Impostor match file names separated by comma")
    ap.add_argument("-g", "--genuine_match_file_names", required=True,
                    help="Genuine match file names separated by comma")
    ap.add_argument("-e", "--experiment_names", required=True,
                    help="Experiment names separated by comma")
    ap.add_argument("-lw", "--line_width", required=False, default=3,
                    help="The width of the plotted curves (default=5)")
    ap.add_argument("-ls", "--legend_font_size", required=False, default=15,
                    help="The size of the legend font (default=20)")
    ap.add_argument("-ht", "--hist", required=False, action='store_true',
                    help="Indicates that the impostor file is in histogram format")
    ap.add_argument("-s", "--save_plots", required=False, action='store_true',
                    help="Indicates whether to save the plots instead of showing them")
    ap.add_argument("-sp", "--save_path", required=False, default='',
                    help="Path to save the plots in the cases where the option -s was specified")
    ap.add_argument("-sf", "--save_format", required=False, default='png',
                    help="Format to save the plots in the cases where the option -s was"
                         " specified. Valid formats are: (png, pdf, ps, eps and svg)")
    ap.add_argument("-sr", "--save_dpi", required=False, default=None,
                    help="Plots resolution (dots per inch) in the cases where the option -s was"
                         " specified. If not given it will default to the value"
                         " savefig.dpi in the matplotlibrc file")
    ap.add_argument("-ts", "--thr_step", required=False, default=0,
                    help="The value in which increase the threshold at each step,"
                         " if 0 (default) we will use the scores as thresholds")
    args = ap.parse_args()

    # Parsing arguments
    genuine_match_file_names = args.genuine_match_file_names.split(',')
    impostor_match_file_names = args.impostor_match_file_names.split(',')
    experiment_names = args.experiment_names.split(',')
    match_pairs = zip(genuine_match_file_names, impostor_match_file_names,
                      experiment_names)
    line_width = int(args.line_width)
    legend_font = int(args.legend_font_size)

    # Preparing plots
    eer_fig = plt.figure()
    eer_plot = eer_fig.add_subplot(111)
    eer_plot.grid(True)
    eer_plot.set_ylabel('Error')
    eer_plot.set_xlabel('Matching Scores')
    eer_plot.set_title('FMR and FNMR Curves')

    det_fig = plt.figure()
    det_plot = det_fig.add_subplot(111)
    det_plot.set_title('DET Curves')
    det_plot.grid(True)
    det_plot.set_ylabel('FNMR')
    det_plot.set_xlabel('FMR')

    roc_fig = plt.figure()
    roc_plot = roc_fig.add_subplot(111)
    roc_plot.set_title('ROC Curves')
    roc_plot.grid(True)
    roc_plot.set_ylabel('1 - FNMR')
    roc_plot.set_xlabel('FMR')
    roc_plot.plot([0, 1], [0, 1], 'k--', linewidth=line_width)

    for match in match_pairs:
        genuine_match_file = join(args.path, match[0])
        impostor_match_file = join(args.path, match[1])
        exp_name = match[2]

        print('Loading genuines file...')
        genuine_match = [__get_score(line) for line in open(genuine_match_file)]

        print('Loading impostor file...')
        impostor_match = [__get_score(line) for line in open(impostor_match_file)]

        print('Calculating probabilities...')
        if args.thr_step != 0 or args.hist:
            # Setting threshold step for step by step calculation
            thr_step = (1 if args.hist and args.thr_step == 0
                        else float(args.thr_step))

            # Calculating probabilities step by step
            roc_info = calculate_eer_step_by_step(genuine_match, impostor_match,
                                                  thr_step, args.hist)
        else:
            # Calculating probabilities using scores as thresholds
            roc_info = calculate_eer(genuine_match, impostor_match)

        # Unboxing probability rates and info
        (thresholds, false_match_rate, false_non_match_rate, eer) = roc_info

        # Printing EER and operation points values
        print(exp_name + ' EER = ' + str(eer))

        index = np.argmin(abs(false_match_rate - 0))
        print(exp_name + ' FNMR_0 = ' + str(false_non_match_rate[index]))

        index = np.argmin(abs(false_match_rate - 0.2))
        print(exp_name + ' FNMR_5 = ' + str(false_non_match_rate[index]))

        index = np.argmin(abs(false_match_rate - 0.1))
        print(exp_name + ' FNMR_10 = ' + str(false_non_match_rate[index]))

        index = np.argmin(abs(false_match_rate - 0.05))
        print(exp_name + ' FNMR_20 = ' + str(false_non_match_rate[index]))

        index = np.argmin(abs(false_match_rate - 0.001))
        print(exp_name + ' FNMR_100 = ' + str(false_non_match_rate[index]))

        index = np.argmin(abs(false_match_rate - 0.0001))
        print(exp_name + ' FNMR_1000 = ' + str(false_non_match_rate[index]))

        print('Ploting Curves...')

        # Plotting FMR and FNMR curves
        eer_plot.plot(thresholds, false_match_rate, label=exp_name + '(FMR)', linewidth=line_width)
        eer_plot.plot(thresholds, false_non_match_rate, label=exp_name + '(FNMR)', linewidth=line_width)

        # Plotting DET Curves
        det_plot.plot(false_match_rate, false_non_match_rate, label=exp_name, linewidth=line_width)

        # Plotting ROC Curves
        roc_plot.plot(false_match_rate, 1 - false_non_match_rate, label=exp_name, linewidth=line_width)

    # Finalizing plots
    eer_plot.legend(loc='best', prop=font.FontProperties(size=legend_font))
    det_plot.legend(loc='best', prop=font.FontProperties(size=legend_font))
    roc_plot.legend(loc='best', prop=font.FontProperties(size=legend_font))

    # Showing plots or saving plots
    if args.save_plots:
        # Parsing dpi
        dpi = None if args.save_dpi is None else int(args.save_dpi)

        # saving plots
        eer_fig.savefig(join(args.save_path, 'EER' + '.' + args.save_format), dpi=dpi)
        det_fig.savefig(join(args.save_path, 'DET' + '.' + args.save_format), dpi=dpi)
        roc_fig.savefig(join(args.save_path, 'ROC' + '.' + args.save_format), dpi=dpi)

        # closing plots
        plt.close()
    else:
        plt.show()



# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

# Get the long description from the README file
with open('README.rst') as f:
    readme = f.read()

setup(
    # Package name
    name='pyeer',

    # Package version
    version='0.2.1',

    entry_points={
        'console_scripts': [
            'geteerinf = pyeer.eer_info:get_eer_info',
        ],
    },

    include_package_data=True,

    package_data={
        # If any package contains *.txt or *.rst files, include them:
        'pyeer': ['example_files/hist/*.txt', 'example_files/non_hist/*.txt'],
    },

    # Included packages
    packages=find_packages(),

    # Package author information
    author=u'Manuel Aguado Martínez',
    author_email='manuelaguadomtz@gmail.com',
    url='https://www.researchgate.net/profile/Manuel_Aguado_Martinez2',

    # Package requirements
    install_requires=['numpy',
                      'matplotlib'],

    # Package description
    description='A python package with utilities to calculate Equal Error Values,'
                ' operation points and to plot the probability error curves.',
    long_description=readme,
    keywords=['Equal Error Rate', 'False Matching Rate', 'False Non-matching rate', 'ROC', 'DET', 'EER', 'FMR', 'FNMR'],

)